//=======================================================================================
//
//  Purpose:    Implement the basics of a ShiftRegister class using SPI to output data
//              to a series of 74HC595 shift registers.
//
//  Copyright (C) 2011 Mark Stevens
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software") to use
//  the Software in a non-commercial environment subject to the following conditions:
//
//  1.  The above copyright notice and this permission notice shall be included in
//      all copies or substantial portions of the Software.
//  2.  The original author shall be acknowledged in any software and documentation
//      including (but not restricted to) books, articles and online posts.
//
//  Use in a commercial environment is subject to the written consent of the author
//  and/or copyright holder.
//
//  The Software, documentation, instructions for use and associated materials
//  are provided "as is" without warranty of any kind express or implied.  The
//  author and copyright holders shall ihn no way be held liable for any claim
//  or damages arising as a result of the use of the software and/or documentation.
//
//=======================================================================================
using System;
using Microsoft.SPOT.Hardware;

namespace Nevyn.co.uk
{
    partial class ShiftRegister
    {
        #region Private Variables

        /// <summary>
        /// Array containing the bits to be output to the shift register.
        /// </summary>
        private bool[] _bits = null;

        /// <summary>
        /// Number of chips required to implement this ShiftRegister.
        /// </summary>
        private static int _numberOfChips;

        /// <summary>
        /// SPI interface used to communicate with the shift registers.
        /// </summary>
        private static SPI _spi = null;

        #endregion

        #region Constructor(s)

        /// <summary>
        /// Default constructor.
        /// </summary>
        private ShiftRegister()
        {
        }

        /// <summary>
        /// Constructor for the ShiftRegister.
        /// </summary>
        /// <param name="bits">Number of bits in the shift register (should be a multiple of 8 bits).</param>
        public ShiftRegister(int bits, Cpu.Pin latchPin = Cpu.Pin.GPIO_Pin8, SPI.SPI_module spiModule = SPI.SPI_module.SPI1, uint speedKHz = 10)
        {
            if ((bits > 0) && ((bits % 8) == 0))
            {
                _bits = new bool[bits];
                _numberOfChips = bits / 8;
                for (int index = 0; index < bits; index++)
                {
                    _bits[index] = false;
                }

                if (_spi == null)
                {

                    SPI.Configuration config;

                    config = new SPI.Configuration(SPI_mod: spiModule, ChipSelect_Port: latchPin,
                                                   ChipSelect_ActiveState: false, ChipSelect_SetupTime: 0,
                                                   ChipSelect_HoldTime: 0, Clock_IdleState: true, Clock_Edge: false,
                                                   Clock_RateKHz: speedKHz);

                    _spi = new SPI(config);
                }
            }
            else
            {
                throw new ArgumentOutOfRangeException("ShiftRegister: Size must be greater than zero and a multiple of 8 bits");
            }
        }

        #endregion

        #region Methods

        /// <summary>
        /// Overload the index operator to allow the user to get/set a particular 
        /// bit in the shift register.
        /// </summary>
        /// <param name="bit">Bit number to get/set.</param>
        /// <returns>Value in the specified bit.</returns>
        public bool this[int bit]
        {
            get
            {
                if ((bit >= 0) && (bit < _bits.Length))
                {
                    return (_bits[bit]);
                }
                throw new IndexOutOfRangeException("ShiftRegister: Bit index out of range.");
            }
            set
            {
                if ((bit >= 0) && (bit < _bits.Length))
                {
                    _bits[bit] = value;
                }
                else
                {
                    throw new IndexOutOfRangeException("ShiftRegister: Bit index out of range.");
                }
            }
        }

        /// <summary>
        /// Send the data to the SPI interface.
        /// </summary>
        public void LatchData()
        {
            byte[] data = new byte[_numberOfChips];

            for (int chip = 0; chip < _numberOfChips; chip++)
            {
                data[chip] = 0;
                byte bitValue = 1;
                int offset = chip * 8;
                for (int bit = 0; bit < 8; bit++)
                {
                    if (_bits[offset + bit])
                    {
                        data[chip] |= bitValue;
                    }
                    bitValue <<= 1;
                }
            }
            _spi.Write(data);
        }

#if DEBUG
        /// <summary>
        /// Convert the output to a string.
        /// </summary>
        /// <returns>String representation of the ShiftRegister.</returns>
        public string ToString()
        {
            string result = "";

            for (int index = _bits.Length; index > 0; index--)
            {
                if ((index % 4) == 0)
                {
                    result += " ";
                }
                if (_bits[index - 1])
                {
                    result += "1";
                }
                else
                {
                    result += "0";
                }
            }
            return (result);
        }
#endif

        #endregion
    }
}
